@@ -1,15 +1,15 @@
 /*****************************************************************************
- * Secure Locate v2.4                                                        *
- * November 28, 2000                                                         *
+ * Secure Locate v2.5                                                        *
+ * December 30, 2000                                                         *
  * Programmed by: Kevin Lindsay                                              *
- * Copyright (c) 1994, 1999, 2000 Kevin Lindsay & Netnation Communications,  *
- *                                Inc. & Free Software Foundation, Inc.      *
+ * Copyright (c) 1999, 2000 Kevin Lindsay & Netnation Communications Inc. &  *
+ *                          James A. Woods <jwoods@adobe.com>                *
  *                                                                           *
  * NetNation Communications   http://www.netnation.com/                      *
- * Secure Locate: ftp://ftp.linux-geek.org/slocate/                          *
+ * Secure Locate: ftp://ftp.geekreview.org/slocate/                          *
  *                ftp://ftp.mkintraweb.com/pub/linux/slocate/                *
  *                                                                           *
- *           Web: http://www.linux-geek.org/slocate/                         *
+ *           Web: http://www.geekreview.org/slocate/                         *
  *                                                                           *
  * Patches:  Sean Mcnulty <lazy@ccf.rutgers.edu>                             *
  *                * Fixed a bug which caused the decode function to fail.    *
@@ -40,9 +40,6 @@
  *                                                                           *
  * Report any Bugs to: klindsay@mkintraweb.com                               *
  *                                                                           *
- * Even though Secure Locate was written from scratch, some recognition goes *
- * to James A. Woods <jwoods@adobe.com> for writing the original GNU locate. *
- *                                                                           *
  * This product includes software developed by the University of             *
  * California, Berkeley and its contributors.                                *
  *                                                                           *
@@ -83,8 +80,13 @@
 #include <ctype.h>
 #include <fnmatch.h>
 #include <regex.h>
+#include "config.h"
+
+#ifdef HAVE_FTS_H
 #include <fts.h>
-//#include "sl_fts.h"
+#else
+#include "sl_fts.h"
+#endif
 
 #include "link.h"
 #include "misc.h"
@@ -96,11 +98,13 @@
 # undef FNM_CASEFOLD
 #endif
 
+
 /* GLOBALS */
 
-#define VERSION "Secure Locate v2.4 - Released November 28, 2000\n"
+#define SL_RELEASE "December 30, 2000"
+#define SL_VERSION "Secure Locate " VERSION " - Released " SL_RELEASE "\n"
 #define GRPFILE "/etc/group"
-#define SLOC_ESC 0x80
+#define SLOC_ESC -0x80
 #define SLOC_GRP "slocate"
 #define SLOC_UID 0
 #define MIN_BLK 4096
@@ -147,6 +151,10 @@ int first=1;
 char slevel='1';
 unsigned short SLOC_GID=65534; /* Seems to be the default nobody gid */
 int max_queries=0;
+int ADD_SLOCATEDB = 1;
+
+#define WARNING 0
+#define FATAL 1
 
 int decode_db(char *database, char *str);
 
@@ -157,14 +165,14 @@ void
 {
 	int i;
 	printf("%s\n"
-	       "Copyright (c) 1994, 1999, 2000 Kevin Lindsay & Netnation Communications Inc. &\n"
-	       "                               Free Software Foundation, Inc.\n\n"
+	       "Copyright (c) 1999, 2000 Kevin Lindsay & Netnation Communications Inc. &\n"
+	       "James A. Woods <jwoods@adobe.com>\n\n"
 	       "search usage:   %s [-qi] [-d <path>] [--database=<path>] <search string>...\n"
 	       "                %s [-r <regexp>] [--regexp=<regexp>]\n"
 	       "database usage: %s [-qv] [-o <file>] [--output=<file>]\n"
 	       "                %s [-e <dir1,dir2,...>] [-f <fs_type1,...> ] [-l <level>]\n"
-	       ,VERSION,progname,progname,progname,progname);
-	
+	       ,SL_VERSION,progname,progname,progname,progname);
+
 	for (i = 0; i < strlen(progname)-1; i+=1)
 	    printf(" ");
 	
@@ -212,11 +220,11 @@ void
 	       "\n"
 	       "Author: Kevin Lindsay\n"
 	       "Bugs:   klindsay@mkintraweb.com\n"	       
-	       "FTP:    ftp://ftp.linux-geek.org/slocate/\n"
+	       "FTP:    ftp://ftp.geekreview.org/slocate/\n"
 	       "        ftp://ftp.mkintraweb.com/pub/linux/slocate/\n"
-	       "HTTP:   http://www.linux-geek.org/slocate/\n"
+	       "HTTP:   http://www.geekreview.org/slocate/\n"
 	       "\n");
-	
+
 	exit(0);
 }
 
@@ -254,27 +262,29 @@ int
     validate_db(char *database)
 {
 	int fd;
+	int myerrno;
 	char buf[2];
 	
 	fd = open(database,O_RDONLY);
 	
 	if (fd == -1)
-	    return 0;
+	    return errno;
 	
 	if (read(fd,buf,2) < 2) {
+		myerrno = errno;
 		close(fd);
-		return 0;
+		return myerrno;
 	}
 	
 	if ((buf[0] != '1' && buf[0] != '0') ||
 	    buf[1] != 0) {
 		close(fd);
-		return 0;
+		return -1;
 	}
 	
 	close(fd);
-	
-	return 1;
+
+	return 0;
 }
 
 /* Parse Output Path */
@@ -287,24 +297,28 @@ void
 	if (!path || strlen(path) == 0) return;
 	
 	ret_val = creat(path, S_IRWXU|S_IRGRP);
-	if(ret_val==-1)
-	{
-		if (!QUIET)
-		    fprintf(stderr,"%s: could not create database: %s: %s\n", progname,path,strerror(errno));
-		exit(1);
-	}
+	if(ret_val==-1)	
+	    report_error(FATAL,QUIET,"%s: parse_create_path: could not create database: %s: %s\n", progname,path,strerror(errno));
 	
 	SLOCATEDB = malloc(strlen(path)+1);
+	if (!SLOCATEDB)
+	    report_error(FATAL,QUIET,"%s: parse_create_path: 'SLOCATEDB': malloc: %s\n",progname,strerror(errno));
 	SLOCATEDB[0] = '\0';
 	strcat(SLOCATEDB, path);
 	
 	TMPSLOCATEDB = malloc(strlen(path)+5);
+	if (!TMPSLOCATEDB)
+	    report_error(FATAL,QUIET,"%s: parse_create_path: 'TMPSLOCATEDB': malloc: %s\n",progname,strerror(errno));
+	
 	TMPSLOCATEDB[0] = '\0';
 	strcat(TMPSLOCATEDB, path);
 	strcat(TMPSLOCATEDB, ".tmp");
 }
 
-/* Parse Path */
+/* Parse Database Paths.
+ * 
+ * Concatenate all database paths into one string and validate each database.
+ */
 
 void
     parse_decode_path(char *path)
@@ -312,28 +326,49 @@ void
 	char *pathcopy;
 	char *part;
 	int i;
+	int res_errno;
 	
+	/* Make sure path is not empty */
 	if (!path || strlen(path) == 0) return;
 	
+	/* Check how many paths are currently in the string. */
 	i = 1;
 	part = path;
 	while ((part = strchr(part+1, ':'))) i++;
-	SLOCATE_PATH = malloc(i * sizeof(char *));
 	
-	i = 0;
+	/* Allocate enough space to fit existing paths plus new one */
+	SLOCATE_PATH = malloc(i * sizeof(char *));
+	if (!SLOCATE_PATH)
+	    report_error(FATAL,QUIET,"%s: parse_decode_path: 'SLOCATE_PATH': malloc: %s\n",progname,strerror(errno));
+	
 	pathcopy = malloc(strlen(path)+1);
-	pathcopy[0] = '\0';
-	strcat(pathcopy, path);
+	if (!pathcopy)
+	    report_error(FATAL,QUIET,"%s: parse_decode_path: 'pathcopy': malloc: %s\n",progname,strerror(errno));
+
+	strcpy(pathcopy,path);
+	
+	/* Get first path just incase a path with ':'s were specified*/
 	part = strtok(pathcopy, ":");
 	
+	i = 0;
+	/* Loop through all paths and validate them.
+	 * use SLOCATE_PATH as an array of pointers to each
+	 * path */
 	while (part) {
-		if (validate_db(part))
+		/* Make sure the path is valid */
+		if (!(res_errno = validate_db(part)))
 		    SLOCATE_PATH[i++] = part;
-		else if (!QUIET)
-		    fprintf(stderr,"%s: this is not a valid slocate database: %s\n",progname,part);
+		else {
+			if (res_errno == -1)
+			    report_error(WARNING,QUIET,"%s: this is not a valid slocate database: %s\n",progname,part);
+			else
+			    report_error(WARNING,QUIET,"%s: could not open database: %s: %s\n",progname,part,strerror(res_errno));
+		}
+		/* Get next path */
 		part = strtok(NULL, ":");
 	}
-	
+
+	/* Null terminate array */
 	SLOCATE_PATH[i] = NULL;
 }
 
@@ -354,7 +389,7 @@ void
 	    usage();
 	
 	if (!strcmp(option,"VERSION")) {
-		printf("%s",VERSION);
+		printf("%s",SL_VERSION);
 		exit(0);
 	}
 	
@@ -371,11 +406,15 @@ void
 	
 	if (!strcmp(option,"DATABASE")) {
 		parse_decode_path(ptr);
+		ADD_SLOCATEDB = 0;
 	}
 	
 	if (!strcmp(option,"REGEXP")) {
 		REGEXP = 1;
 		regexp = malloc(strlen(ptr)+1);
+		if (!regexp)
+		    report_error(FATAL,QUIET,"%s: parse_dash: 'regexp': malloc: %s\n",progname,strerror(errno));
+
 		*regexp = 0;
 		strcat(regexp,ptr);
 		while ((database = SLOCATE_PATH[i++])) res |= decode_db(database, ptr);
@@ -387,19 +426,13 @@ void
 
 /* Make sure that the DB Directory exists */
 
-int
-    check_dir()
+void
+    check_dir(char *dir_str)
 {
-	int res=1;
 	struct stat tmpstat;
-	
-	if (lstat(SLOCATEDB_DIR,&tmpstat) == -1) {
-		if (!QUIET)
-		    fprintf(stderr,"%s: error accessing DB Directory: %s : %s\n",progname,SLOCATEDB_DIR,strerror(errno));
-		res=0;
-	}
-	
-	return(res);
+		
+	if (!dir_str || lstat(dir_str,&tmpstat) == -1)
+	    report_error(FATAL,QUIET,"%s: error accessing DB Directory: %s : %s\n",progname,SLOCATEDB_DIR,strerror(errno));
 }
 
 /* Get the GID for group slocate */
@@ -411,11 +444,8 @@ unsigned short
 	struct group *grpres;
 	
 	if ((grpres = getgrnam(SLOC_GRP)) == NULL) {
-		if (!QUIET) {
-			fprintf(stderr,"%s: Could not find the group: %s in the /etc/group file.\n",progname,SLOC_GRP);
-			fprintf(stderr,"This is a result of the group missing or a corrupted group file.\n");
-		}
-		exit(1);
+		report_error(WARNING,QUIET,"%s: Could not find the group: %s in the /etc/group file.\n",progname,SLOC_GRP);
+		report_error(FATAL,QUIET,"This is a result of the group missing or a corrupted group file.\n");
 	}
 	
 	GGID = grpres->gr_gid;
@@ -437,6 +467,9 @@ int
 	EXCLUDE=1;   
 	
 	excludestr = (char *)malloc(1);
+	if (!excludestr)
+	    report_error(FATAL,QUIET,"%s: parse_exclude: 'excludestr': malloc: %s\n",progname,strerror(errno));
+	
 	excludestr[0] = '\0';
 	
 	elen = strlen(estr);
@@ -457,6 +490,8 @@ int
 		}
 		
 		excludestr = realloc(excludestr,grow);
+		if (!excludestr)
+		    report_error(FATAL,QUIET,"%s: parse_exclude: 'excludestr': realloc: %s\n",progname,strerror(errno));
 		strcat(excludestr,"*");
 		strncat(excludestr,ptr1,ptr2-ptr1);
 		strcat(excludestr,"*");
@@ -470,9 +505,13 @@ int
 	
 	if (!EXCLUDE_DIR) {
 		EXCLUDE_DIR = malloc(strlen(excludestr)+1);
+		if (!EXCLUDE_DIR)
+		    report_error(FATAL,QUIET,"%s: parse_exclude: 'EXCLUDE_DIR': malloc: %s\n",progname,strerror(errno));
 		*EXCLUDE_DIR = 0;      
 	} else {
 		EXCLUDE_DIR = realloc(EXCLUDE_DIR,strlen(EXCLUDE_DIR)+strlen(excludestr)+1);
+		if (!EXCLUDE_DIR)
+		    report_error(FATAL,QUIET,"%s: parse_exclude: 'EXCLUDE_DIR': realloc: %s\n",progname,strerror(errno));
 	}
 	
 	strcat(EXCLUDE_DIR,excludestr);
@@ -546,6 +585,8 @@ int
 	if (estr) {
 		estr_size = strlen(estr)+2;
 		newestr = malloc(estr_size+1);
+		if (!newestr)
+		    report_error(FATAL,QUIET,"%s: parse_fs_exclude: 'newestr': malloc: %s\n",progname,strerror(errno));
 		*newestr = 0;
 		strcat(newestr,"*");
 		for (ptr = estr; *ptr != 0; ptr += 1) {
@@ -553,6 +594,8 @@ int
 			if (*ptr == ',') {
 				estr_size += 1;
 				newestr = realloc(newestr,estr_size+1);
+				if (!newestr)
+				    report_error(FATAL,QUIET,"%s: parse_fs_exclude: 'newestr': realloc: %s\n",progname,strerror(errno));				
 				strcat(newestr,"*");
 			} else {
 				strncat(newestr,ptr,1);
@@ -566,23 +609,29 @@ int
 	
 #ifdef __FreeBSD__
 	{
-		struct statfs *fs_stat;
+		struct statfs fs_stat;
 		long bufsize=4096;
 		int num_mounts=0;
 		int i;
 		char *exclude_str=NULL;
 		
 		fs_stat = malloc(sizeof(struct statfs));
-		
+		if (!fs_stat)
+		    report_error(FATAL,QUIET,"%s: parse_fs_exclude: 'fs_stat': malloc: %s\n",progname,strerror(errno));
+
 		num_mounts = getfsstat(fs_stat,bufsize,MNT_WAIT);
 		
 		for (i = 0; i < num_mounts; i+=1) {
 			if (strstr(estr,get_fs_type(fs_stat[i].f_type))) {
 				if (!exclude_str) {
 					exclude_str = malloc(strlen(fs_stat[i].f_mntonname)+1);
+					if (!exclude_str)
+					    report_error(FATAL,QUIET,"%s: parse_fs_exclude: 'exclude_str': malloc: %s\n",progname,strerror(errno));
 					*exclude_str = 0;
 				} else {
 					exclude_str = realloc(exclude_str,strlen(exclude_str)+strlen(fs_stat[i].f_mntonname)+2);
+					if (!exclude_str)
+					    report_error(FATAL,QUIET,"%s: parse_fs_exclude: 'exclude_str': realloc: %s\n",progname,strerror(errno));
 					strcat(exclude_str,",");
 				}
 				
@@ -607,9 +656,7 @@ int
 		
 		if (*fbuf == 0) {
 			fbuf += 1;
-			if (!QUIET)
-			    fprintf(stderr,"%s: File System Exclude: Could not open file %s: %s\n",progname,MTAB_FILE,fbuf);
-			exit(1);
+			report_error(FATAL,QUIET,"%s: File System Exclude: Could not open file %s: %s\n",progname,MTAB_FILE,fbuf);
 		}
 		
 		head_ptr = fbuf;
@@ -637,6 +684,8 @@ int
 			/* Check if file sytem type exists in exclude string */
 			
 			match_str = realloc(match_str,strlen(head_ptr)+3);
+			if (!match_str)
+			    report_error(FATAL,QUIET,"%s: parse_fs_exclude: 'match_str': realloc: %s\n",progname,strerror(errno));			
 			match_str[0] = '*';
 			match_str[1] = 0;
 			strcat(match_str,head_ptr);
@@ -654,29 +703,27 @@ int
 				while (*head_ptr != ' ' && head_ptr != fbuf)
 				    head_ptr -= 1;
 				
-				if (head_ptr == fbuf) {
-					if (!QUIET)
-					    fprintf(stderr,"%s: File System Exclude: (1) corrupt mtab file: %s\n",progname,MTAB_FILE);
-					exit(1);                  
-				}
-				
+				if (head_ptr == fbuf)
+				    report_error(FATAL,QUIET,"%s: File System Exclude: (1) corrupt mtab file: %s\n",progname,MTAB_FILE);
+
 				head_ptr += 1;
 				
-				if (!(tail_ptr = strchr(head_ptr,' '))) {
-					if (!QUIET)
-					    fprintf(stderr,"%s: File System Exclude: (2) corrupt mtab file: %s\n",progname,MTAB_FILE);
-					exit(1);
-				}
+				if (!(tail_ptr = strchr(head_ptr,' ')))
+				    report_error(FATAL,QUIET,"%s: File System Exclude: (2) corrupt mtab file: %s\n",progname,MTAB_FILE);
 				
 				*tail_ptr = 0;
 				
 				/* Add the path to the exclude string */
 				if (!exclude_str) {
 					exclude_str = malloc(strlen(head_ptr)+1);
+					if (!exclude_str)
+					    report_error(FATAL,QUIET,"%s: parse_fs_exclude: 'exclude_str': malloc: %s\n",progname,strerror(errno));
 					*exclude_str = 0;
 					strcat(exclude_str,head_ptr);
 				} else {
 					exclude_str = realloc(exclude_str,strlen(exclude_str)+strlen(head_ptr)+2);
+					if (!exclude_str)
+					    report_error(FATAL,QUIET,"%s: parse_fs_exclude: 'exclude_str': realloc: %s\n",progname,strerror(errno));
 					strcat(exclude_str,",");
 					strcat(exclude_str,head_ptr);
 				}
@@ -720,14 +767,11 @@ void
 	fbuf = load_file(UPDATEDB_CONF);
 	
 	if (*fbuf == 0) {
-		fbuf += 1;
-		/*      if (!QUIET) {
-		 fprintf(stderr,"%s: could not access %s: %s\n",progname,UPDATEDB_CONF,fbuf);
-		 fflush(stderr);
-		 } */
+		report_error(WARNING,QUIET,"%s: could not access %s: %s\n",progname,UPDATEDB_CONF,fbuf+1);
+		free(fbuf);
 		return;
 	}
-	
+
 	for (times = 0; times < 2; times += 1) {
 		
 		head_ptr = fbuf;
@@ -774,9 +818,13 @@ void
 						
 						if (!parse_str) {
 							parse_str = malloc(strlen(head_ptr)+1);
+							if (!parse_str)
+							    report_error(FATAL,QUIET,"%s: parse_updatedb_conf: 'parse_str': malloc: %s\n",progname,strerror(errno));
 							*parse_str = 0;
 						} else {
 							parse_str = realloc(parse_str,strlen(parse_str)+strlen(head_ptr)+2);
+							if (!parse_str)
+							    report_error(FATAL,QUIET,"%s: parse_updatedb_conf: 'parse_str': realloc: %s\n",progname,strerror(errno));							
 							strcat(parse_str,",");
 						}
 						
@@ -821,6 +869,8 @@ int
 	    char *newstr;
 	    
 	    newstr = malloc(strlen(path)+strlen(filedir)+3);
+	    if (!newstr)
+		report_error(FATAL,QUIET,"%s: match_exclude: 'newstr': malloc: %s\n",progname,strerror(errno));
 	    newstr[0] = '\0';
 	    strcat(newstr,"*");
 	    strcat(newstr,path);
@@ -838,7 +888,7 @@ int
 /* FRCODE - Incremental Encoding algorithm */
 
 int
-    frcode(FILE *fd,char *dir_path, char *filename)
+    frcode(FILE *fd, char *dir_path, char *filename)
 {
 	int res=0;
 	char *cur_path;
@@ -847,10 +897,11 @@ int
 	char *ptr2;
 	
 	int i;
-	
+
 	cur_path = malloc(strlen(dir_path)+strlen(filename)+1);
-	cur_path[0]='\0';
-	strcat(cur_path,dir_path);
+	if (!cur_path)
+		report_error(FATAL,QUIET,"%s: frcode: 'cur_path': malloc: %s\n",progname,strerror(errno));
+	strcpy(cur_path,dir_path);
 	strcat(cur_path,filename);
 	
 	if (VERBOSE)
@@ -871,22 +922,26 @@ int
 		
 		ptr2 = strchr(ptr1,'\0');
 		new_path=malloc((ptr2-ptr1)+1);
+		if (!new_path)
+		    report_error(FATAL,QUIET,"%s: frcode: 'new_path': malloc: %s\n",progname,strerror(errno));
 		new_path[0] = '\0';
 		strncat(new_path,ptr1,ptr2-ptr1);         
 		
 	} else {
 		new_path = malloc(strlen(cur_path)+1);
-		new_path[0] = '\0';
-		strcat(new_path,cur_path);
+		if (!new_path)
+		    report_error(FATAL,QUIET,"%s: frcode: 'new_path': malloc: %s\n",progname,strerror(errno));
+		strcpy(new_path,cur_path);
 		t_num = strlen(new_path);
 	}
 	
-	if (tmp_path != NULL)
+	if (tmp_path)
 	    free(tmp_path);
 	
 	tmp_path = malloc(strlen(cur_path)+1);
-	tmp_path[0] = '\0';
-	strcat(tmp_path,cur_path);
+	if (!tmp_path)
+	    report_error(FATAL,QUIET,"%s: frcode: 'tmp_path': malloc: %s\n",progname,strerror(errno));	
+	strcpy(tmp_path,cur_path);
 	
 	if (fr_num > 127 || fr_num < -127) {
 		putc (SLOC_ESC, fd);
@@ -921,24 +976,24 @@ int
 	FTSENT *file;
 	int i;
 	char **dirchk;
-	
-	
+
 	if (!NEWOUTPUT && UID != SLOC_UID) {
-		if (!QUIET)
-		    fprintf(stderr,"%s: You are not authorized to create a default slocate database!\n",progname);
-		return(1);
-	} 
-	
+		report_error(FATAL,QUIET,"%s: You are not authorized to create a default slocate database!\n",progname);
+	}
+
 	fd = fopen(TMPSLOCATEDB,"w");
+
+	if (!fd)
+	    report_error(FATAL,QUIET,"%s: create_db: fopen: '%s': %s\n",progname,TMPSLOCATEDB,strerror(errno));
+
 	putc(slevel,fd);
 	
 	if (dirstr) {
 		/* Make sure there are no problems accessing the source directory */
-		if (lstat(dirstr,&statres) == -1) {
-			if (!QUIET)
-			    fprintf(stderr,"%s: lstat: d%s/: %s\n",progname,dirstr,strerror(errno));
-			fclose(fd);
-			return(1);
+		if (lstat(dirstr,&statres) == -1) {			
+			report_error(FATAL,QUIET,"%s: lstat: d%s/: %s\n",progname,dirstr,strerror(errno));
+//			fclose(fd);
+//			return(1);
 		}
 		
 		if (strlen(dirstr) > 1) {
@@ -948,17 +1003,24 @@ int
 	}
 	else {
 		dirstr = malloc(2);
+		if (!dirstr)
+		    report_error(FATAL,QUIET,"%s: create_db: 'dirstr': malloc: %s\n",progname,strerror(errno));
 		*dirstr = 0;
 		strcat(dirstr,"/");
 	}
 	
 	dirchk = malloc(sizeof(char **)*2);
-	
+	if (!dirchk)
+	    report_error(FATAL,QUIET,"%s: create_db: 'dirchk': malloc: %s\n",progname,strerror(errno));	
 	*dirchk = dirstr;
 	dirchk[1] = NULL;
 	
-	dir = fts_open(dirchk,FTS_PHYSICAL,NULL);   
+	dir = fts_open(dirchk,FTS_PHYSICAL,NULL);
 	
+	/* If fts_open failes, report and exit */
+	if (!dir)
+	    report_error(FATAL,QUIET,"%s: fts_open: %s\n",progname,strerror(errno));
+
 	/* The new FTS() funtionality */
 	
 	for (i = 0; i > -1; i+=1) {
@@ -981,10 +1043,18 @@ int
 	fts_close(dir);
 	
 	fclose(fd);
-	chdir(prog_CWD);
-	rename(TMPSLOCATEDB,SLOCATEDB);
-	chmod(SLOCATEDB,00640);
-	chown(SLOCATEDB,UID,SLOC_GID);   
+	if (chdir(prog_CWD) == -1)
+	    report_error(FATAL,QUIET,"%s: create_db(): chdir: %s\n",progname,strerror(errno));
+	
+	if (rename(TMPSLOCATEDB,SLOCATEDB) == -1)
+	    report_error(FATAL,QUIET,"%s: create_db(): rename: %s\n",progname,strerror(errno));
+	if (chmod(SLOCATEDB,00640) == -1)
+	    report_error(FATAL,QUIET,"%s: create_db(): chmod: %s\n",progname,strerror(errno));
+	/* Only change group to SLOC_GID if UID == 0 so that regular users can make their
+	   own databases */
+	if (UID == 0 && chown(SLOCATEDB,UID,SLOC_GID) == -1)
+	    report_error(FATAL,QUIET,"%s: create_db(): chown: %s\n",progname,strerror(errno));
+	
 	return res;
 }
 
@@ -1033,30 +1103,28 @@ int
 	char *bucket_of_holding=NULL;
 	
 	if ((fd = open(database,O_RDONLY)) == -1) {
-		if (!QUIET)
-		    fprintf(stderr,"%s: decode_db(): %s: %s\n",progname,database,strerror(errno));
+		report_error(WARNING,QUIET,"%s: decode_db(): %s: %s\n",progname,database,strerror(errno));
 		return(0);
 	}
 	
 	lstat(database,&statres);
 	
 	if (S_ISDIR(statres.st_mode)) {
-		if (!QUIET)
-		    fprintf(stderr,"%s: decode_db(): %s is a directory\n",progname,database);
+		report_error(WARNING,QUIET,"%s: decode_db(): %s is a directory\n",progname,database);
 		return(0);
 	}
 	
 	time(&now);
 	if (now - statres.st_mtime > WARN_SECONDS)
-	{
-		fprintf(stderr,"%s: warning: database %s' is more than %s old\n",progname,database,WARN_MESSAGE);
-	}   
+	    report_error(WARNING,QUIET,"%s: warning: database %s' is more than %s old\n",progname,database,WARN_MESSAGE);
 
 //	slevel = getc(fd);
 	read(fd,one_char,1);
 	slevel = *one_char;
 
 	codedpath = malloc(MIN_BLK);
+	if (!codedpath)
+	    report_error(FATAL,QUIET,"%s: decode_db: 'codedpath': malloc: %s\n",progname,strerror(errno));
 	*codedpath = 0;
 	code_ptr = codedpath;
 
@@ -1067,11 +1135,11 @@ int
 	
 	if (REGEXP) {   
 		preg = malloc(sizeof(regex_t));
+		if (!preg)
+		    report_error(FATAL,QUIET,"%s: decode_db: 'preg': malloc: %s\n",progname,strerror(errno));
 		if ((reg_res = regcomp(preg,regexp,NOCASE?REG_ICASE:0)) != 0) {
 			regerror(reg_res, preg, errbuf,1024);
-			if (!QUIET)
-			    fprintf(stderr,"error: %s: regular expression: %s\n",progname,errbuf);
-			exit(1);
+			report_error(FATAL,QUIET,"error: %s: regular expression: %s\n",progname,errbuf);
 		}
 #ifndef FNM_CASEFOLD
 		
@@ -1084,6 +1152,8 @@ int
 	}
 	
 	bucket_of_holding = malloc(MIN_BLK);
+	if (!bucket_of_holding)
+	    report_error(FATAL,QUIET,"%s: decode_db: 'bucket_of_holding': malloc: %s\n",progname,strerror(errno));	
 	*bucket_of_holding = 0;
 	begin_ptr = bucket_of_holding;
 	tot_size = MIN_BLK;
@@ -1097,6 +1167,8 @@ int
 			    tot_size <<= 1;
 			begin_offset = begin_ptr - bucket_of_holding;
 			bucket_of_holding = realloc(bucket_of_holding,tot_size);
+			if (!bucket_of_holding)
+			    report_error(FATAL,QUIET,"%s: decode_db: 'bucket_of_holding': realloc: %s\n",progname,strerror(errno));
 			begin_ptr = bucket_of_holding + begin_offset;
 		}
 		
@@ -1106,12 +1178,12 @@ int
 
 		cur_size += bytes;
 
-		code_num = *begin_ptr;
+		code_num = (short)*begin_ptr;
 		begin_ptr += 1;
 
-		if (code_num == SLOC_ESC)
-		    code_num = get_short(&begin_ptr);
-		else if (code_num > 127)
+		if (code_num == SLOC_ESC) {
+			code_num = get_short(&begin_ptr);
+		} else if (code_num > 127)
 		    code_num = code_num - 256;
 
 		code_ptr += code_num;
@@ -1123,11 +1195,16 @@ int
 			ch = *begin_ptr;
 			begin_ptr++;
 			pathlen++;
+		
+			if (pathlen < 0)
+				report_error(FATAL,QUIET,"%s: decode_db: 'pathlen == %d'! Corrupt Database!\n",progname,pathlen);
 
 			if (pathlen > code_tot_size) {
 				code_tot_size = pathlen * 2;
 				ptr_offset = code_ptr - codedpath;
 				codedpath = realloc(codedpath,code_tot_size);
+				if (!codedpath)
+				    report_error(FATAL,QUIET,"%s: decode_db: 'codedpath': realloc: %s\n",progname,strerror(errno));					
 				code_ptr = codedpath+ptr_offset;
 			}
 
@@ -1138,8 +1215,9 @@ int
 			/* FIXME: Handle if begin_ptr runs past buffer */
 			
 			if (begin_ptr > bucket_of_holding+cur_size-1 && bytes) {
-				fprintf(stderr,"slocate fluky bug found!\n");
-				fprintf(stderr,"Ack! This shouldn't happen unless you have a path over 4096!\n");
+				fprintf(stderr,"slocate fluky bug found.\n");
+				fprintf(stderr,"Ack! This shouldn't happen unless you have a path over 4096.\n");
+				fprintf(stderr,"This could also be a bogus or corrupt database.\n");
 				fprintf(stderr,"Report this as a bug to klindsay@mkintraweb.com\n");
 				exit(1);
 			}
@@ -1189,19 +1267,7 @@ int
 		if (foundit) {
 			if (slevel == '1') {
 				if (UID == 0 ||  access(codedpath, F_OK) == 0) {
-			        //if (UID == 0 || lstat(codedpath,&statres) == 0) {
 					printit = 1;
-/*					chk1 = codedpath+1;
-					while ((chk1 = strchr(chk1,'/')) != NULL) {
-						chk1++;
-						tmpch=*chk1;
-						*chk1='\0';
-						
-						if (access(codedpath, R_OK) == -1)
-						    printit=0;
-
-						*chk1=tmpch;
-					} */
 				}
 			} else
 			    printit=1;
@@ -1245,100 +1311,124 @@ int
 	char *database;
 	int i=0;
 	
+	/* Get the name of the program as it was executed */
+	progname = ((p = strrchr(argv[0],'/')) ? p+1 : *argv);
 	
-	progname = ((p = strrchr(argv[0],'/')) ? p+1 : *argv);     
-	
+	/* If slocate has been run with the name 'updatedb' then we want to initial
+	 * the variables necessary to update the database instead of search it */
 	if (!strcmp(progname,"updatedb")) {
 		ROOTDIR=1;
+
 #ifndef __FreeBSD__
 		parse_updatedb_conf();
 #endif
+		/* Check to make sure default database directory exists */
+		check_dir(SLOCATEDB_DIR);
 	}
-	
-	/* Get Current Working Directory */
-	
+
+	/* Get Current Working Directory */	
 	getcwd(prog_CWD, 4095);
 	
-	if (strlen(prog_CWD) == 4095) {
-		fprintf(stderr,"%s: Current Working Directory is too large!",progname);
-		exit(1);
-	}
-	
+	/* If the working directory is too long, then report a fatal error */
+	if (strlen(prog_CWD) == 4095)
+	    report_error(FATAL,QUIET,"%s: Current Working Directory is too large!",progname);
 	
+	/* Print usage if not enough arguments were specified */
 	if (args < 2 && !ROOTDIR)
 	    usage();
 	
-	if (!check_dir())
-	    return(1);
-	
+	/* Get user IDs */
 	UID = getuid();        
 	GID = getgid();
-	
-	parse_decode_path(SLOCATEDB);
+
+	/* Add the LOCATE_PATH environment variable to the list of databases to search in */
 	parse_decode_path(getenv("LOCATE_PATH"));
 	
+	/* Loop through specified command line arguments */
 	while ((ch = getopt(args,argv,"VvuhqU:r:o:e:l:d:-:n:f:ci")) != EOF) {
 		switch(ch) {
+		 /* Help */
 		 case 'h':
 			usage();
 			break;
+		 /* Quiet Mode. Don't print warnings or errors. */
 		 case 'q':
 			QUIET=1;
 			break;
+		 /* Print version */
 		 case 'V':
-			printf("%s",VERSION);
+			printf("%s",SL_VERSION);
 			exit(0);
 			break;
+		 /* Turn VERBOSE mode ON */
 		 case 'v':
 			VERBOSE=1;
 			break;
+		 /* Exclude specified directories from database */
 		 case 'e':
 			parse_exclude(optarg);
 			break;
+		 /* Exclude specified filesystems from database */
 		 case 'f':
 			parse_fs_exclude(optarg);
 			break;
+		 /* Parse the updatedb.conf file */
 		 case 'c':
 #ifndef __FreeBSD__
 			parse_updatedb_conf();
 #endif
 			break;
+		 /* Set the security level of the database when creating or updating the
+		  * database. If set to 0, security checks will not be preformed */
 		 case 'l':
 			slevel = optarg[0];
-			if (slevel != '0' && slevel != '1') {
-				if (!QUIET)
-				    fprintf(stderr,"%s: ERROR: Security level must be 0 or 1.\n",progname);
-				exit(1);
-			}
+			if (slevel != '0' && slevel != '1')
+			    report_error(FATAL,QUIET,"%s: ERROR: Security level must be 0 or 1.\n",progname);
 			break;
+		 /* Update the default database */
 		 case 'u':
+			/* Check to make sure default database directory exists */
+			check_dir(SLOCATEDB_DIR);
 			ROOTDIR=1;
+			ADD_SLOCATEDB = 0;
 			break;
+		 /* Update a specified database */
 		 case 'U':
+			ADD_SLOCATEDB = 0;
 			ROOTDIR=0;
 			SPECDIR=1;
 			spec_dir = malloc(strlen(optarg)+1);
-			*spec_dir = 0;         
+			if (!spec_dir)
+			    report_error(FATAL,QUIET,"%s: main: 'spec_dir': malloc: %s\n",progname,strerror(errno));
+			*spec_dir = 0;
 			strcat(spec_dir,optarg);
 			break;
+		 /* Search database with specified regular expression */
 		 case 'r':
 			REGEXP = 1;
 			regexp = malloc(strlen(optarg)+1);
+			if (!regexp)
+			    report_error(FATAL,QUIET,"%s: main: 'regexp': malloc: %s\n",progname,strerror(errno));
 			*regexp = 0;
 			strcat(regexp,optarg);
 			while ((database = SLOCATE_PATH[i++])) res |= decode_db(database, optarg);
 			return(res);
 			break;
-		 case 'o':
+		 /* Specify the database path to write to when creating or updating a database */
+		 case 'o':			
 			parse_create_path(optarg);
 			NEWOUTPUT=1;
 			break;
+		 /* Specify the database to search in */
 		 case 'd':
 			parse_decode_path(optarg);
+			ADD_SLOCATEDB = 0;
 			break;
+		 /* Parse --<argument> type arguments */
 		 case '-':
 			parse_dash(optarg);
 			break;
+		 /* Limit the amount of search results */
 		 case 'n':
 			for (i=0; i < strlen(optarg); i+=1) {
 				if (!isdigit(optarg[i])) {
@@ -1348,6 +1438,7 @@ int
 			}
 			max_queries = atoi(optarg);
 			break;
+		 /* Make search case insensitive */
 		 case 'i':
 			NOCASE=1;         
 			break;
@@ -1357,13 +1448,25 @@ int
 			
 		}
 	}
-	
+
+	/* Only add the default database path if necessary.  We don't want todo this
+	 * when creating/updating a database or when a database has been specified
+	 * with the -d option, etc */
+
+	if (ADD_SLOCATEDB)
+	    parse_decode_path(SLOCATEDB);
+
+	/* Get the 'slocate' group GID */
 	SLOC_GID = get_gid(GRPFILE);
 	
-	if (SPECDIR) {
-		res = create_db(spec_dir);
-	} else if (ROOTDIR)
-		    res = create_db((char *)NULL);
+	
+	/* if the -U option has been used, start to create the database at specified path */
+	if (SPECDIR)
+	    res = create_db(spec_dir);
+	/* Create databse from default '/' path */
+	else if (ROOTDIR)
+	    res = create_db((char *)NULL);
+	/* search the database(s) */
 	else if (argv[optind]) {
 		int j = optind;
 		int i = 0;
@@ -1377,8 +1480,10 @@ int
 			    res |= decode_db(database, argv[j]);
 			i = 0; j += 1;
 			
-		}
-	} else
+		}		
+	}
+	/* Print usage */
+	else
 	    usage();
 	
 	return(res);
